#from bisect import bisect_right,bisect_left
import math
import numpy as np
#import random
from friendly_core import FriendlyCore,min_rho_for_FC
import matplotlib.pyplot as plt    
from dist_utils import l2_dist_squared   
from time import sleep         

'''
noisy_sigma: computes a private estimation of noise that needs to be added for (rho,delta)-zCDP
 in the insertion/deletion model (where neighboring databases are according to insertion/deletion model, i.e., D and D'=D\cup{z}).
'''
def noisy_sigma(n: int, rho: float, delta: float, r: float):
    # rho1 is set to the maximum of the function f(x) = (n-sqrt(ln(1/delta)/x))*sqrt(rho-x). (internal page 356)
    rho1 = (math.sqrt(math.log(1/delta))*rho/n)**(2/3)
    rho2 = rho-rho1/(1-delta)
    print("noisy_sigma(n=%d,rho=%lf)->(rho1=%lf,rho2=%lf)" % (n,rho,rho1,rho2))
    if rho1>=rho:
        print("**** noisy_sigma: rho1 is too large. setting to 0.5*rho")
        rho1 = 0.5*rho*(1-delta)
        rho2 = 0.5*rho
    
    # hn denotes hat{n} - a private estimation of n.
    hn = n - math.sqrt(math.log(1/delta)/rho1) - 1 + math.sqrt(1/(2*rho1))*np.random.randn() 
    if n==0 or hn<=0:
        print("n=%d, hn = %lf, are too small" % (n,hn))
        raise RuntimeError('n is too small')
    return 2*r/(hn*math.sqrt(2*rho2))

'''
FriendlyAvg: f-friendly (rho,delta)-zCDP algorithm, where f(x,y)=1 <=> l2_dist(x,y)<=r

Inputs:
D: Database of points in R^d
d: Dimension
rho, delta: Privacy parameters.
r: Diameter.

'''
def FriendlyAvg(D: np.ndarray, d: int, rho: float, delta: float, r: float):
    n = len(D)
    sigma = noisy_sigma(n,rho,delta,r)
    ha = np.average(np.array(D), axis=0) + np.random.normal(0, sigma, size=d)
    return ha

####################################### Private Averaging ################################################


########################### Known Diameter ###############################
'''
FC_Avg_KnownDiam: (rho,delta)-zCDP algorithm that has averaging utility only for database of diameter r.

Inputs:
D: Database of points in R^d.
d: Dimension.
rho, delta: Privacy parameters.
r: Diameter.

'''
def FC_Avg_KnownDiam(D: np.ndarray, d: int, rho: float, delta: float, r: float):
    n = len(D)
    rho1 = min_rho_for_FC(n,delta/2) 
    rho2 = rho - rho1
    print("FC_Avg_KnownDiam: n = %d, rho1 = %lf, rho2 = %lf, rho = %lf" % (n,rho1,rho2,rho))
    if rho1 >= rho:
        print("**** FC_Avg_KnownDiam: rho1 is too large. setting to 0.75*rho")
        rho1 = 0.75*rho
        rho2 = rho-rho1
    
    r_squared = r*r
    def dist_r(x: np.ndarray, y: np.ndarray):
        if (l2_dist_squared(x,y) <= r_squared):
            return 1
        return 0
    C = FriendlyCore(D,dist_r,rho1,delta/2)
    return FriendlyAvg(C,d,rho2,delta/2,r)
    

'''
For the averaging experiments, 
FC_Avg wraps FC_Avg_KnownDiam for the case that D consists of samples from N(0,I)
(and therefore the expected diameter r can easily be computed in advance).
''' 
def FC_Avg(D: np.ndarray, d: int, rho: float, delta: float):
    n = len(D)
    r = (d**0.5 + (math.log(100*n))**0.5)*(2**0.5)
    return FC_Avg_KnownDiam(D,d,rho,delta,r)


########################### Unknown Diameter ###############################

'''
CheckDiam: (rho,delta)-zCDP algorithm that outputs a boolean value 
of whether most of the points in D belong to a ball of diameter r, or not. 

Inputs:
D: Database of points in R^d, or ordered k-tuples over R^d.
d: Dimension
rho, delta: Privacy parameters.
r: Diameter.
dist_square_func: measure the l2_dist squared distance between two points,
or the maximal squared distance (per coordinate) of ordered k-tuples.
m: sample size (to speed-up running time).
'''
def CheckDiam(D: np.ndarray, rho: float, beta: float, r: float, dist_square_func, m: int = 200):
    n = len(D)
    a = 0.0
    r_squared = r**2
    
    
    S = D
    if n>2*m:
        print("CheckDiam: Using a sample of size %d to speed-up running time" % m)
        pi = [i for i in range(n)]
        np.random.shuffle(pi)
        S = np.array([D[pi[i]] for i in range(m)])
    else:
        print("CheckDiam: Using O(n^2) computation.")
    sample_size = len(S)
    for x in D:
        for y in S:
            if dist_square_func(x,y)<=r_squared:
                a = a + 1/sample_size
    
    # ha denotes hat{a} - private estimation of a.  
    ha = a + np.random.randn()*math.sqrt(2/rho)
    gap = math.sqrt(4*math.log(1/beta)/rho)
    print("CheckDiam:  n = %d, r = %lf, a = %d, gap = %lf, ha = %lf" % (n,r,a,gap,ha))
    if ha >= n - gap:
        if a<0.8*n:
            print("**** CheckDiam: Success on a BAD diameter.") 
        return True
    else:
        if a>=0.99*n:
            print("**** CheckDiam: Failure on a GOOD diameter.")
        return False

'''
FindDiam: (rho,delta)-zCDP algorithm that outputs a diameter r that estimates the effective diameter of D.

Inputs:
D: Database of points in R^d, or ordered k-tuples over R^d.
d: Dimension
rho, delta: Privacy parameters.
r_min, r_max: lower and upper bounds (respectively) of the effective diameter.
dist_square_func: measure the l2_dist squared distance between two points,
or the maximal squared distance (per coordinate) of ordered k-tuples. 
'''
    
def FindDiam(D,rho,beta,r_min,r_max, dist_square_func):
    base = 2
    t = math.ceil(math.log(r_max/r_min,base))
    ell = math.ceil(math.log2(t))
    
    low = 0
    high = t
    
    # Searching the effective diameter via a private binary search
    while low <= high:
        
        mid = (low + high)//2
        x = math.pow(base,mid)
        r=x*r_min
        result = CheckDiam(D,rho/ell,beta/ell,r, dist_square_func)
        
        if not result:
            low = mid + 1
        elif result:
            high = mid - 1
    
    r = math.pow(base,low)*r_min
    print("FindDiam: r = %lf is chosen" % r)    
    return r



'''
FC_Avg_UnknownDiam: (rho,delta)-zCDP algorithm that estimates the average of the points in D
when the effective diameter is unknown (only rough bounds are given).

Inputs:
D: Database of points in R^d.
d: Dimension.
rho, delta: Privacy parameters.
beta: utility parameter
r_min, r_max: Lower and upper bounds (respectively) of the effective diameter.
'''
def FC_Avg_UnknownDiam(D: np.ndarray, d: int, rho: float, delta: float, beta: float, r_min: float, r_max: float):
    n = len(D)
    t = math.ceil(math.log(r_max/r_min,1.5))
    rho1 = min(rho/2, 800*math.log(1/beta)*math.log2(t)/(n**2))
    rho2 = rho-rho1
    print("FC_Avg_UnknownDiam: rho1=%lf, rho2=%lf" % (rho1,rho2))
    r = FindDiam(D,rho1,beta,r_min,r_max,l2_dist_squared)
    return FC_Avg_KnownDiam(D,d,rho2,delta,r)



#####################################################################
