import numpy as np
import math
from friendly_core import FriendlyCore,min_rho_for_FC
from dist_utils import l2_dist_squared,max_l2_dist_squared
from avg_algos import FC_Avg_UnknownDiam,FC_Avg_KnownDiam, noisy_sigma, FindDiam
from clustering_utils import non_private_kMeans_Clustering_3
import matplotlib.pyplot as plt

__show_plots__ = 0


'''
Given a database of k-tuples, the function returns a database of all the points in all the k-tuples.
'''
def Points(T: np.ndarray):
    PT = []
    for X in T:
        for x in X:
            PT.append(x)
    return np.array(PT)
  

'''
CreateTuples: A method that given a database of points S, creates a database of k-tuples.
This is done by splitting S into random pieces, and applying a non-private clustering algorithm in each pieace for obtaiting a k-tuple.

Inputs:
S: Database of points in R^d.
k: Clustering parameter.
num_of_tuples: The number of pieces that are used (and the number of resulting k-tuples).
non_private_alg: A non-private clustering algorithm that given database X and parameter k as inputs, outputs k centers (k-tuple).
'''
def CreateTuples(S: np.ndarray, k: int, num_of_tuples: int, non_private_alg):
    n = len(S)
    pi = [i for i in range(n)]
    np.random.shuffle(pi)
    
    m = n//num_of_tuples
    T = []
    for i in range(num_of_tuples):
        X = np.array([S[pi[j]] for j in range(m*i,m*(i+1))])
        centers = non_private_alg(X,k)
        if (i+1)%10 == 0:
            print("[FriendlyCore] Created tuple %d" % (i+1))
        T.append(centers)
        
    print("[FriendlyCore] num_of_tuples = %d" % len(T))
    return T


######################################### Ordered Tuples ##################################################

'''
FriendlyAvgOrdTup: f-friendly (rho,delta)-zCDP algorithm for averaging ordered k-tuples, 
where for two tuples X=(x1,...,xk) and Y=(y1,...,yk): f(X,Y)=1 <=> l2_dist(xi,yi)<=r forall i.

Inputs:
D: Database of ordered k-tuples over R^d.
d: Dimension.
k: Tuple size.
rho,delta: Privacy parameters.
r: Maximal diameter (per i)
'''
def FriendlyAvgOrdTup(D: np.ndarray, d: int, k: int, rho: float, delta: float, r: float):
    n = len(D)
    sigma = noisy_sigma(n,rho/k,delta/k,r)
    hA = []
    for j in range(k):
        Q = np.array([X[j] for X in D])
        ha = np.average(Q, axis=0) + np.random.normal(0, sigma, size=d)
        hA.append(ha)
    return hA

'''
FC_AvgOrdTup: (rho,delta)-zCDP algorithm for averaging ordered k-tuples over R^d.
For utility, rough bounds on the effective diameters in each coordinate are required.


Inputs:
D: Database of ordered k-tuples over R^d.
d: Dimension.
k: Tuple size.
rho,delta: Privacy parameters.
beta: Utility parameter.
r_min, r_max: Lower and upper bounds (resp) of the maximal diameter (per i).
'''
def FC_AvgOrdTup(D: np.ndarray, d: int, k: int, rho: float, delta: float, beta: float, r_min: float, r_max: float):
    n = len(D)
    t = math.ceil(math.log(r_max/r_min,1.5))
    rho1 = min(0.2*rho, 400*math.log(1/beta)*math.log2(t)/(n**2))
    rho2 = min_rho_for_FC(n,delta/2)
    rho3 = rho - rho1 - rho2
    if rho3 <= 0:
        print("**** FC_AvgOrdTup: Setting (rho1,rho2,rho3) to (0.3,0.6,0.1)*rho")
        rho1 = 0.3*rho
        rho2 = 0.6*rho
        rho3 = 0.1*rho
    print("FC_AvgOrdTup: rho1 = %lf, rho2 = %lf, rho3 = %lf, rho = %lf" % (rho1,rho2,rho3,rho))
      
    r = FindDiam(D,rho1,beta,r_min,r_max,max_l2_dist_squared)
    r_squared = r**2
    def dist_r(X,Y):
        if (max_l2_dist_squared(X,Y) <= r_squared):
            return 1
        return 0
    C = FriendlyCore(D,dist_r,rho2,delta/2)
    centers = FriendlyAvgOrdTup(C,d,k,rho3,delta/2,r)
    if __show_plots__:
        D0 = np.array([Y[0] for Y in D])
        D1 = np.array([Y[1] for Y in D])
        D2 = np.array([Y[2] for Y in D])
        plt.scatter(D0[:,0],D0[:,1],s=1,color='purple')
        plt.scatter(D1[:,0],D1[:,1],s=1,color='green')
        plt.scatter(D2[:,0],D2[:,1],s=1,color='orange')
        plt.scatter(np.array(centers)[:,0],np.array(centers)[:,1],s=16,color='red')
        plt.xlim(-10,10)
        plt.ylim(-10,20)
        plt.savefig('./plots/clustering_demo/step6.png')
        print("STEP6: the red points are private averages of each set ... ")
        plt.show()
    return centers


#########################################  Unordered Tuples ######################################### 

'''
OrderTuple: Order the coordinates of Y={y_1,...,y_k} according to X={x_1,...,x_k}.
Namely, output ord_Y={y_1',...,y_k'} where for each i: y_i' = y_j for y_j being the closest point to x_i among the points in Y.
If ord_Y is not a permutation of Y, the function raises an exeption. 
'''
def OrderTuple(X: np.ndarray, Y: np.ndarray):
    k = len(X)
    ord_Y = []
    test = []
    for _ in range(k):
        test.append(0)
    for i in range(k):
        j = np.argmin([l2_dist_squared(X[i],y) for y in Y])
        if 1==test[j]:
            #print("**** OrderTuple: Not a permutation")
            raise RuntimeError('OrderTuple has failed')
        else:
            test[j]=1
        ord_Y.append(Y[j])
    return ord_Y


'''
match: a predicate of unoredered k-tuples with parameter gamma.
The predicate checks whether there is a 1-1 maching between the points in X and the points in Y.
The parameter gamma is the required matching quality.
'''
def match(X: np.ndarray, Y: np.ndarray, gamma: float):
    k = len(X)
    try:
        ord_Y = OrderTuple(X,Y)
    except:
        return False
    for i in range(k):
        dist_ii = l2_dist_squared(X[i],ord_Y[i])
        for j in range(k):
            dist_ij = l2_dist_squared(X[i],ord_Y[j])
            dist_ji = l2_dist_squared(X[j],ord_Y[i])
            #print("dist_%d%d = %lf, dist_%d%d = %lf, Delta=%lf" % (i,i,dist_ii,i,j,dist_ij,Delta))
            if j!=i and dist_ii >= gamma**2 * min(dist_ij,dist_ji):
                return False
    return True


'''
FriendlyReorder: An algorithm that order the unordered k-tuples in T with the following guarantee:
If A is (rho,delta)-zCDP, then B(T) = A(FriendlyReorder(T)) is match_{1/7}-friendly (rho,delta)-zCDP.

Inputs:
T: Database of k-tuples over R^d (i.e., each element in T consists of k points in R^d).
'''
def FriendlyReorder(T: np.ndarray):
    n = len(T)
    if n==0: raise RuntimeError('No tuples!')
    X = T[0]
    
    D = np.array([OrderTuple(X,Y) for Y in T])
        
    if __show_plots__:
        PT = Points(T)
        plt.scatter(PT[:,0],PT[:,1],s=1)
        plt.xlim(-10,10)
        plt.ylim(-10,20)
        plt.savefig('./plots/clustering_demo/step3.png')
        print("STEP3: Presenting the points of all the %d-tuples in the CORE (%d points of %d tuples)...." % (len(X), len(D)*len(X), len(D)))
        plt.show()
        plt.scatter(PT[:,0],PT[:,1],s=1)
        plt.scatter(np.array(X)[:,0],np.array(X)[:,1],s=16,color='black')
        plt.xlim(-10,10)
        plt.ylim(-10,20)
        plt.savefig('./plots/clustering_demo/step4.png')
        print("STEP4: Picking an arbitrary %d-tuple (the black one) ..." % (len(X)))
        plt.show()
        D0 = np.array([Y[0] for Y in D])
        D1 = np.array([Y[1] for Y in D])
        D2 = np.array([Y[2] for Y in D])
        plt.scatter(D0[:,0],D0[:,1],s=1,color='purple')
        plt.scatter(D1[:,0],D1[:,1],s=1,color='green')
        plt.scatter(D2[:,0],D2[:,1],s=1,color='orange')
        plt.scatter(np.array(X)[:,0],np.array(X)[:,1],s=16,color='black')
        plt.xlim(-10,10)
        plt.ylim(-10,20)
        plt.savefig('./plots/clustering_demo/step5.png')
        print("STEP5: Partitioning the points of the tuples into %d groups according to the black tuple ...." % (len(X)))
        plt.show()
    
    return D


'''
FC_kTupleClustering: Our main (rho,delta)-zCDP algorithm for k-tuple clustering.
The utility guarantee: If T consists of k-tuples that are ``close enough'' to each other, then the output will be a new k-tuple that is close to them.

Inputs:
T: Database of k-tuples over R^d (i.e., each element in T consists of k points in R^d).
d: Dimension.
k: Num of clusters (and the length of each tuple).
rho,delta: Privacy parameters.
r_min,r_max: Lower and upper bounds (resp) of the maximal diameter of each cluster (required only for utility).
'''
def FC_kTupleClustering(T: np.ndarray, d: int, k: int, rho: float, delta: float, beta: float, r_min: float, r_max: float):
    rho1 = 0.45*rho
    rho2 = 0.55*rho
    def match_17(X: np.ndarray, Y: np.ndarray):
        return match(X,Y,1/7)
    C = FriendlyCore(T,match_17,rho1,delta/2)
    D = FriendlyReorder(C)
    Y = FC_AvgOrdTup(D,d,k,rho2,delta/2,beta/2,r_min,r_max)
    return (C,Y)

#####################################   General Clustering #################################

def Noisy_Lloyd_step(Points: np.ndarray, d: int, centers: np.ndarray, rho: float, delta: float, Lambda: float):
    k = len(centers)
    Clusters = []
    for _ in range(k):
        Clusters = Clusters + [[]]
    for x in Points:
        i = np.argmin([l2_dist_squared(x, c) for c in centers])
        Clusters[i].append(x)
    for i in range(k):
        cluster_i = np.array(Clusters[i])
        n_i = len(cluster_i)
        sigma_i = noisy_sigma(n_i,rho,delta,r=2*Lambda)
        print("[NoisyLloyd] n_%d = %d, sigma_%d = %lf" % (i,n_i,i,sigma_i))
        centers[i] = np.average(cluster_i, axis=0) + np.random.normal(0, sigma_i, size=d) 
    return centers

'''
FC_Clustering: Our main (rho,delta)-zCDP clustering algorithm.
Given a database of points in the ball B(0,Lambda) over R^d, the algorithm either:
(1) Outputs k centers that are close to the output of the non private algorithm (that is given as input), OR
(2) Fail, and exception is thrown. 
Case (1) is expected to happen if the solution of non_private_alg is ``stable enough'' w.r.t. the database D.
That is, when splitting D into num_of_tuples pieces, then non_private_alg outputs almost the same centers in each piece. 
We note that ``stability'' is a function of separability and size: Less separation requires more data. 

Inputs:
D: Database of points in R^d.
d: Dimension.
k: Number of clusters.
rho, delta: Privacy parameters.
beta: Utility parameter.
r_min: Lower bound on the diameter of each cluster (required only for utility).
Lambda: A bound on the L_2 norm of the data points (required only for utility).
num_of_tuples: the number of pieces that D is splitted into (from each piece, a k-tuple is generated).
non_private_alg: The non-private clustering algorithm that is executed in each piece.
use_Lloyd: Whether to use an additional private Lloyd step after the k-tuple clustering.
'''
def FC_Clustering(D: np.ndarray, d: int, k: int, rho: float, delta: float, beta: float, r_min: float, Lambda: float, num_of_tuples: int, non_private_alg = non_private_kMeans_Clustering_3, use_Lloyd : bool = True):
    
    T = CreateTuples(D,k,num_of_tuples,non_private_alg)
    
    if __show_plots__:
        plt.scatter(D[:,0],D[:,1],s=1)
        plt.xlim(-10,10)
        plt.ylim(-10,20)
        plt.savefig('./plots/clustering_demo/step1.png')
        print("STEP1: Presenting all the %d points ...." % (len(D)))
        plt.show()
        PT = Points(T)
        plt.scatter(PT[:,0],PT[:,1],s=1)
        plt.xlim(-10,10)
        plt.ylim(-10,20)
        plt.savefig('./plots/clustering_demo/step2.png')
        print("STEP2: Presenting the points of all the %d-tuples (%d points of %d tuples)...." % (k, k*num_of_tuples, num_of_tuples))
        plt.show()

    rho1 = rho
    if use_Lloyd:    
        rho1 = 2.5*min_rho_for_FC(num_of_tuples,delta/2)
        print("FC_Clustering: rho1=%lf" % rho1)
        if rho1 >= rho:
            rho1 = 0.9*rho
            print("*** FC_Clustering: rho1 is too large. Setting to 0.9*rho=%lf" % rho1)
    rho2 = rho-rho1
    (C,centers) = FC_kTupleClustering(T,d,k,rho1,delta/2,beta,r_min,r_max=2*Lambda)
    if __show_plots__:
        PT = Points(C)
        plt.scatter(np.array(PT)[:,0],np.array(PT)[:,1],s=1)
        plt.xlim(-10,10)
        plt.ylim(-10,20)
        plt.scatter(np.array(centers)[:,0],np.array(centers)[:,1],s=16,color='red')
        plt.savefig('./plots/clustering_demo/step7.png')
        print("STEP7: removing the partitioning colors ... ")
        plt.show()
        plt.scatter(D[:,0],D[:,1],s=1)
        plt.xlim(-10,10)
        plt.ylim(-10,20)
        plt.scatter(np.array(centers)[:,0],np.array(centers)[:,1],s=16,color='red')
        plt.savefig('./plots/clustering_demo/step8.png')
        print("STEP8: the private red points located in the original database ... ")
        plt.show()
    if use_Lloyd:
        print("[FC_Clustering] Performing a final Noisy Lloyd step")
        centers = Noisy_Lloyd_step(D,d,centers,rho2,delta/2,Lambda)
        if __show_plots__:
            plt.scatter(D[:,0],D[:,1],s=1)
            plt.xlim(-10,10)
            plt.ylim(-10,20)
            plt.scatter(np.array(centers)[:,0],np.array(centers)[:,1],s=16,color='yellow')
            print("STEP9: the yellow points are the result of a private Lloyd step ... ")
            plt.savefig('./plots/clustering_demo/step9.png')
            plt.show()

    return centers
    
    
##########################################################            