import math
import numpy as np
from time import sleep

def min_rho_for_FC(n: int, delta: float, m: int = 300):
    # Estimation of a good (minimal) rho for FriendlyCore.
    if n>m:
        # With a sample (for speed-up running time)
        return 8*math.log(n/(8*delta))/n
    else:
        # Without a sample.
        return 4*math.log(n/(8*delta))/n


def zCDPFilter(D: np.ndarray, f, rho: float, delta: float, m: int):
    n = len(D)
    xi = 0
    S = D
    if n>m:
        print("zCDPFilter: Using a sample of size %d to speed-up running time." % m)
        pi = [i for i in range(n)]
        np.random.shuffle(pi)
        S = np.array([D[pi[i]] for i in range(m)])
        # When using a sample, we split delta, half for the sample (in xi, the threshold parameter), and half for the rest of the algorithm.
        xi = math.sqrt(math.log(2*n/delta)/(2*m))
        delta = delta/2
    else:
        print("zCDPFilter: Using the O(n^2) computation.")

    # rho1 is set to the (approx) minimum of the function f(x) = (n+sqrt(ln(1/delta)/x))*sqrt(rho-x). (internal page 357)
    rho1 = math.log(2*n/delta)**(1/3) * (2*rho/(4*n))**(2/3)
    rho2 = rho - rho1
    if rho1 >= rho:
        #print("**** zCDPFilter: rho1 is too large. setting to 0.5*rho")
        rho1 = 0.5*rho
        rho2 = 0.5*rho
    print("[FriendlyCore] rho = %lf: Splitted into rho1 = %lf, rho2 = %lf" % (rho, rho1,rho2))    
    
    # hn denotes hat{n} - the private estimation of n
    hn = n + math.sqrt(math.log(2/delta)/rho1) + math.sqrt(1/(2*rho1))*np.random.randn()
    V = []

    for i in range(n):
        w_i = np.sum([f(D[i],y) for y in S])*(len(D)/len(S))
        z_i = w_i-(1/2+xi)*n
        if i%(n//20)==0:
            print("zCDPFilter: Friends of element %d: %d%%" % (i,100*w_i/n))
            #print("zCDPFilter: w_%d = %d" % (i,w_i))
        hz_i = z_i + math.sqrt(hn/(8*rho2))*np.random.randn()
        v_i = (hz_i >= math.sqrt(hn*math.log(2*hn/delta)/(4*rho2)) + 0.5)
        if w_i>=0.99*n and v_i==0:
            print("**** zCDPFilter: Good element has been removed (%d%% friends)" % (100*w_i/n))
        V.append(v_i)
    return V


"""
FriendlyCore Algorithm.

Inputs:

D: Database
f: Predicate of the form f(x,y)->{0,1}
rho, delta: Privacy parameters.
m: Sample size, for improving running time to O(n*m) instead of O(n^2).

"""
def FriendlyCore(D: np.ndarray, f, rho: float, delta: float, m: int = 300):
    n = len(D)
    V = zCDPFilter(D,f,rho,delta,m)
    C = np.array([D[i] for i in range(n) if V[i]==1])
    print("[FriendlyCore] Core size = %d" % len(C))
    return C
